% SchererKose_RobFiltEx
%
% This file duplicates the results in the Scherer/Kose paper
% using our robust filter design code.
% 
% Reference: 
% Robustness with dynamic IQCs: An exact state-space characterization
% of nominal stability with applications to robust estimation
% by C.W. Scherer and I.E. Kose, Automatica, vol 44, p. 1666-1675, 2008.
%
% PJS 6/16/2010  Initial Coding

%------------------------------------------------------------------------
% Problem Data (Section 5 of reference)
% Create P such that
%   [v;y;z] = P*[w;d]
% Punc wraps w = Delta*v around P, i.e.
%   [y;z] = Punc(Delta)*d   where Punc = lft(Delta,P)
%
% Note 1: The Automatica paper has a typo in the "B" matrix multiplying p.
% In fact, this is the "B" matrix multiplying [p;w].  The correct data is 
% provided in the CDC paper.
%
% Note 2: The ex. in the ref. uses a repeated real uncertainty delta*I_2
% to model the uncertainty in c. In fact, c enters into the A matrix as:
%     A = Anom + delta*A1
% where A1 = 1.5*blkdiag(zeros(2),[ -1 1; 2 -2])
% A1 only has rank 1 and hence only one copy of the uncertainty is 
% needed in this example. The robust control toolbox auto-reduces
% the Delta to one copy and Punc (below) will show only 1 occurrence.
%------------------------------------------------------------------------
if 0
    % Form given in the reference
    % B1 = [0;0;-1.5;3]*[1 -1]  (B1 is rank 1)
    del = ureal('del',0);
    Delta = del*eye(2);

    A = [0 0 1 0; 0 0 0 1; -2 1 -2 2; 2 -2 4 -4];

    B1 = [zeros(2,2); -1.5 1.5; 3 -3];
    B2 = zeros(4,2);
    B2(3,1) = 1;
    B = [B1 B2];

    C1 = [zeros(2) eye(2)];
    C2 = [1 0 0 0];
    C3 = [0 1 0 0];
    C = [C1;C2;C3];

    D = zeros(4,4);
    D(3,4) = 1;
else
    % Equivalent form using only one copy of del    
    del = ureal('del',0);
    Delta = del;
    
    A = [0 0 1 0; 0 0 0 1; -2 1 -2 2; 2 -2 4 -4];

    B1 = [0; 0; -1.5; 3];
    B2 = zeros(4,2);
    B2(3,1) = 1;
    B = [B1 B2];

    C1 = [0 0 1 -1];
    C2 = [1 0 0 0];
    C3 = [0 1 0 0];
    C = [C1;C2;C3];

    D = zeros(3,3);
    D(2,3) = 1;    
end
P = ss(A,B,C,D);
Punc = lft(Delta,P);

%------------------------------------------------------------------------
% Optimal gain from Scherer/Kose:  Robust L2 filter computed using their 
% LMI formulation. Basis functions are used for the unc. scalings.
%------------------------------------------------------------------------
fprintf('Scherer/Kose: Minimal L2-L2 gain = %4.3f \n\n',2.64);

%------------------------------------------------------------------------
% Use freq. gridding to compute lower bound on optimal L2-L2 gain 
%------------------------------------------------------------------------
nomega = 250;
omega = logspace(-1,1,nomega);
nmeas = 1;
tic
[gain1,Fopt1,PIopt1] = robfiltlb(Punc,nmeas,omega);
t1=toc;
[gain1lb,mxidx1]=max(gain1);
fprintf('Freq. Grid: Lower Bound on minimal L2-L2 gain = %4.3f\n',gain1lb);
fprintf('Comp time = %4.3f sec\n',t1);

% Plot lower bnd vs. frequency
figure(1)
ph=semilogx(omega,gain1);
xl=xlabel('Frequency (rad/sec)');
yl=ylabel('Lower Bound');
tl = title('Performance Lower Bound vs. Frequency');
grid on;
set([xl yl tl gca],'FontSize',18);
set(ph,'LineWidth',3);

% Verify that IQC condition is satisfied at all frequencies
Nz = 1; Nw = 1; Nd = 2; Ny = nmeas;
Puncfr = frd(Punc,omega);
[Pfr,Delta] = lftdata(Puncfr.ResponseData);
[Nw,Nv]=size(Delta);
lmax = zeros(nomega,1);
for i1=1:nomega
    % Get blocks of P
    P1 = Pfr(1:Nv,:,i1);
    P1I = [P1; eye(Nw) zeros(Nw,Nd)];
    P2 = Pfr(Nv+1:Nv+Ny,:,i1);
    P3 = Pfr(Nv+Ny+1:end,:,i1);    
    
    FPI = [-Fopt1(:,:,i1)*P2+P3; zeros(Nd,Nw) eye(Nd)];
    M = blkdiag( eye(Nz), -gain1(i1)^2*eye(Nd) );
            
    IQCLMI = P1I'*PIopt1(:,:,i1)*P1I + FPI'*M*FPI;
    lmax(i1) = max(real(eig(IQCLMI)));       
end
fprintf(['Max eval from IQC condition = ' num2str(max(lmax))]);
fprintf('  (should be < 0) \n\n');

%------------------------------------------------------------------------
% Compute a lower bound by gridding on the uncertainty space
% (This method is used in papers/thesis by K. Sun and Packard)
%------------------------------------------------------------------------
nunc = 20;
ncon = 1;
delgrid = linspace(-1,1,nunc);
gain2 = zeros(nunc,1);
for i1=1:nunc
    Pd = usubs(Punc,'del',delgrid(i1));
    M = [Pd(2,:) -1; Pd(1,:) 0];
    [K,CL,GAM,INFO] = hinfsyn(M,nmeas,ncon);
    gain2(i1) = GAM;
end
gain2lb = max(gain2);
fprintf('Unc. Grid: Lower Bound on minimal L2-L2 gain = %4.3f\n\n',gain2lb);

%------------------------------------------------------------------------
% Compute upper bound using DKSYN
%------------------------------------------------------------------------

% Scale plant by d-->e gain to get mu approximately equal to  1
% Note: DKYSN default treats delta as a complex parameter uncertainty.
% This is conservative, but turning on the Mixed Mu code actually does
% worse (larger value for optimal mu)
Munc = [Punc(2,:) -1; Punc(1,:) 0];
scl = diag([0.2054 1]); 
opt = dkitopt;
%opt = dkitopt('MixedMU','on');  % Mixed mu does worse than complex mu
tic
[F,CL,GAM,DKINFO] =dksyn(scl*Munc,nmeas,ncon,opt);
t2 = toc;
gain3ub = 1/scl(1,1);

% Use WCGAIN to compute gain achieved by DKSYN filter
% WCGAIN treats delta as a real parameter uncertainty so gain4ub should
% be <= gain3ub.
GF = [-F 1]*Punc;
GFfr = frd(GF,omega);
[MAXGAIN4,MAXGAINUNC4,WCINFO4] = wcgain(GFfr);
gain4ub = MAXGAIN4.UpperBound;
fprintf('DKSYN: L2-L2 gain = %4.3f \n',gain4ub);
fprintf('Comp time = %4.3f sec\n\n',t2);

%------------------------------------------------------------------------
% Compute upper bound using ROBFILTSYN
%------------------------------------------------------------------------
fprintf('Beginning Frequency-Gridding Robust Filter Synthesis \n');
Fbasis = {tf(1,[1 0.1]) tf(1,[1 1]), tf(1,[1 10])};
Fbasis = [Fbasis {tf(1) tf(1,[1 2*0.8*0.1 0.64])}];
tic
[gopt,Fopt5,PIopt,xopt] = robfiltsyn(Punc,Fbasis,nmeas,omega);
t3=toc;
gain5ub = gopt;
fprintf('Freq. Grid Syn: L2-L2 gain = %4.3f \n',gain5ub);
fprintf('Comp time = %4.3f sec\n',t3);

% Use WCGAIN to compute gain achieved by DKSYN filter
% Should get gain5ub = gain6ub;
GF = [-Fopt5 1]*Punc;
GFfr = frd(GF,omega);
[MAXGAIN6,MAXGAINUNC6,INFO6] = wcgain(GFfr);
gain6ub = MAXGAIN6.UpperBound;
fprintf('WCGAIN: L2-L2 gain for freq. grid filter = %4.3f \n',gain6ub);
fprintf(' (Two gains above should be equal)\n\n');

% Plot Bode mag of optimal filter
[mm,pp] = bode(Fopt5,omega);
mm = 20*log10(mm(:));
pp = pp(:);

figure(2);
subplot(211)
ph1=semilogx(omega,mm);
xl1=xlabel('Frequency (rad/sec)');
yl1=ylabel('|F(j\omega)|   (dB)');
tl1 = title('Bode Plot of Optimal Filter');
ah1 = gca;
grid on;

subplot(212)
ph2=semilogx(omega,mm);
xl2=xlabel('Frequency (rad/sec)');
yl2=ylabel('\angle F(j\omega)   (deg)');
ah2 = gca;

grid on;
set([xl1 yl1 tl1 xl2 yl2 ah1 ah2],'FontSize',18);
set([ph1 ph2],'LineWidth',3);
