function [gain,PI,y,ev] = rfsolve(P,Fbasisfr,omega,cidx,blk,dim);
% Solve robust filter design problem with constraints enforced on a finite 
% set of frequencies.  The optimal IQC scalings computed on omega(cidx) 
% are then linearly interpolated to compute scalings at all frequencies 
% in omega. 
%
% INPUTS
% P, Fbasisfr: Response data for P and the F basis functions
% omega: Frequency vector
% cidx: The constraints are enforced at the frequencies of omega specified 
%    by the indices in cidx.
% blk: N-by-2 array specifying the uncertainty block in the old MU
%    notation.
% dim: Structure containing the signal dimensions for the problem.
%
% OUTPUTS
% gain: Optimal gain computed on omega(cidx)
% PI: Optimal IQC scalings.  These are optimized on omega(cidx) and
%   then linearly interpolated to compute PI on all omega.
% y: Nbasis-by-1 cell array of matrices that specify the optimal
%      linear combination of basis functions. 
% ev: The maximum real eigenvalue of the IQC LMI at each frequency.

% Initialize outputs
gain = [];
Ffr = [];
PI = [];
y = [];
ev = [];

% Unpack dimensions
ny = dim.ny;
nz = dim.nz;
ne = dim.ne;
nd = dim.nd;
nv = dim.nv;
nw = dim.nw;

nbasis = length(Fbasisfr);
nblk = size(blk,1);
ncidx = length(cidx);
nomega = length(omega);


% Construct structure of LMI variables for uncertainty scalings
setlmis([]);
PIr = cell(2,2,ncidx);
PIi = cell(2,2,ncidx);
PI11 = cell(ncidx);
vars = 0;
for i1=1:ncidx
    PI11r = [];  PI11i = [];
    PI12r = [];  PI12i = [];
    PI22r = [];  PI22i = [];
    for i2=1:nblk
        if blk(i2,2) == 0 % repeated real & repeated complex
            blksz = abs(blk(i2,1));
            Xr  = symdec(blksz,vars);
            vars = vars+blksz*(blksz+1)/2;
            Xi = skewdec(blksz,vars);
            vars = vars+blksz*(blksz-1)/2;

            if blk(i2,1)<0  % repeated real
                Yr = skewdec(blksz,vars);
                vars = vars+blksz*(blksz-1)/2;
                Yi  = symdec(blksz,vars);
                vars = vars+blksz*(blksz+1)/2;
            else % repeated complex
                Yr = zeros(blksz);
                Yi = zeros(blksz);
            end
            PI11r = blkdiag(PI11r,Xr);
            PI11i = blkdiag(PI11i,Xi);

            PI22r = blkdiag(PI22r,-Xr);
            PI22i = blkdiag(PI22i,-Xi);

            PI12r = blkdiag(PI12r,Yr);
            PI12i = blkdiag(PI12i,Yi);

        else   % full block
            blksz = blk(i2,:);
            X1r = (vars+1)*eye(blksz(1));
            X1i = zeros(blksz(1));

            X2r = (vars+1)*eye(blksz(2));
            X2i = zeros(blksz(2));

            Yr = zeros( blksz(2), blksz(1) );
            Yi = zeros( blksz(2), blksz(1) );

            PI11r = blkdiag(PI11r,X2r);
            PI11i = blkdiag(PI11i,X2i);

            PI22r = blkdiag(PI22r,-X1r);
            PI22i = blkdiag(PI22i,-X1i);

            PI12r = blkdiag(PI12r,Yr);
            PI12i = blkdiag(PI12i,Yi);

            vars = vars+1;
        end
    end

    % Create variables for LMI scalings, filter, and gain
    PIr{i1} = lmivar(3, [PI11r PI12r; PI12r' PI22r] );
    PIi{i1} = lmivar(3, [PI11i PI12i; -PI12i' PI22i] );
    PI11{i1} = lmivar( 3 , [PI11r PI11i; -PI11i PI11r] );
end

% LMI vars for filter basis functions and gain
ybasis = cell(nbasis,1);
for i1 = 1:nbasis;
    [ybasis{i1},cnt,junk] = lmivar( 2 , [nz ny] );
    vars = vars+cnt;
end

gsq = lmivar(3, blkdiag(zeros(nw),(vars+1)*eye(nd)) );
vars = vars+1;


% Create LMI
cnt = 1;
F = cell(nbasis,1);
for i1=1:ncidx
    % Get blocks of P and F basis functions at frequency cidx(i1)
    P1 = P(1:nv,:, cidx(i1) );    
    P1I = [P1; eye(nw) zeros(nw,nd)];    
    P2 = P(nv+1:nv+ny,:, cidx(i1) );
    P3 = P(nv+ny+1:end,:, cidx(i1) );
    for i2 = 1:nbasis
        F{i2} = Fbasisfr{i2}(:,:, cidx(i1) );
    end

    % Construct main LMI for constant filter synthesis
    
    % REAL(L) in (1,1)
    lmiterm([cnt 1 1 PIr{i1}],real(P1I)',real(P1I));
    lmiterm([cnt 1 1 PIr{i1}],imag(P1I)',imag(P1I));
    lmiterm([cnt 1 1 PIi{i1}],imag(P1I)',real(P1I),'s');    
    lmiterm([-cnt 1 1 gsq],1,1);
    for i2 = 1:nbasis
        lmiterm([cnt 2 1 ybasis{i2}],1,real(F{i2}*P2));
    end
    lmiterm([-cnt 2 1 0],real(P3));
    
    lmiterm([cnt 2 2 0],-eye(nz));
    
    % REAL(L) in (2,2)
    lmiblk = [0 2 2 0];
    lmiterm(lmiblk+[cnt 1 1 PIr{i1}],real(P1I)',real(P1I));
    lmiterm(lmiblk+[cnt 1 1 PIr{i1}],imag(P1I)',imag(P1I));
    lmiterm(lmiblk+[cnt 1 1 PIi{i1}],imag(P1I)',real(P1I),'s');    
    lmiterm(lmiblk+[-cnt 1 1 gsq],1,1);
    for i2 = 1:nbasis
        lmiterm(lmiblk+[cnt 2 1 ybasis{i2}],1,real(F{i2}*P2));
    end
    lmiterm(lmiblk+[-cnt 2 1 0],real(P3));
    
    lmiterm(lmiblk+[cnt 2 2 0],-eye(nz));

    % IMAG(L) in (2,1)
    lmiblk = [0 2 0 0];
    lmiterm(lmiblk+[cnt 1 1 PIr{i1}],real(P1I)',imag(P1I));
    lmiterm(lmiblk+[cnt 1 1 PIr{i1}],-imag(P1I)',real(P1I));
    lmiterm(lmiblk+[cnt 1 1 PIi{i1}],real(P1I)',real(P1I));
    lmiterm(lmiblk+[cnt 1 1 PIi{i1}],imag(P1I)',imag(P1I));
    
    for i2 = 1:nbasis
        lmiterm(lmiblk+[cnt 2 1 ybasis{i2}],1,imag(F{i2}*P2));
        lmiterm(lmiblk+[-cnt 1 2 -ybasis{i2}],imag(F{i2}*P2)',1);
    end
    lmiterm(lmiblk+[-cnt 2 1 0],imag(P3));
    lmiterm(lmiblk+[-cnt 1 2 0],-imag(P3)');
    
    % Add LMI for PI11> 0
    % (Play around with tolerance, xtol*I, if numerical issues arise)
    lmiterm([-(cnt+1) 1 1 PI11{i1}],1,1);
    %xtol = 1e-5;       
    %lmiterm([2 1 1 0],xtol*eye(2*nv));

    % Increment LMI counter
    cnt = cnt+2;
end

%fprintf('# of LMI vars = %d \t',vars);

% Solve LMI
lmisys = getlmis;
opts = [0 0 0 0 1];
obj = zeros(vars,1);
obj(end) = 1;
%tic
if 1
    % Use LMILab MINCX
    [copt,xopt] = mincx(lmisys,obj,opts);
else
    % Try with Sedumi using the LMILab translator
    opts = [];
    opts.prt = 0;
    [copt,xopt,Zopt,info] = minsdp('sedumi',lmisys,obj,opts);
    copt = copt(1);
end
%fprintf(' comptime = %4.3f \t',toc);
gain = sqrt(copt);

% Get matrix coefficients that specify optimal filter
for i1 = 1:nbasis;
    y{i1} = dec2mat(lmisys,xopt,ybasis{i1});
end

% Use xopt to form PI at frequencies specified by cidx
PI = zeros([(nv+nw) (nv+nw) nomega]);
i = sqrt(-1);
for i1=1:ncidx
    PI(:,:,cidx(i1)) = dec2mat(lmisys,xopt,PIr{i1}) + i*dec2mat(lmisys,xopt,PIi{i1});
end
    
% Use linear interpolation to form PI at other frequencies 
% PI is held constant for indices outside of [cidx(1),cidx(end)]
didx = setdiff(1:nomega,cidx);
for i1=1:length(didx)
    % Find nearest previous/next frequencies in cidx list
    ptr = didx(i1);
    prev = max( cidx(cidx<ptr) );
    if isempty(prev)
        prev = ptr;
    end
    next = min( cidx(cidx>ptr) );
    if isempty(next)
        next = ptr;
    end

    % Perform linear interpolation
    scl = ( omega(ptr)-omega(prev) ) / ( omega(next)-omega(prev) );
    PI(:,:,ptr) = (1-scl)*PI(:,:,prev) + scl*PI(:,:,next);        
end    


% Check IQC LMI condition on dense grid 
% Note: Back off slightly on gain
gtol = 1e-3;
gain = (1+gtol)*gain;

ev = zeros(nomega,1);
Ffr = zeros(nz,ny,nomega);
M = blkdiag( eye(nz), -gain^2*eye(nd) );
for i1=1:nomega
    % Get blocks of P
    P1 = P(1:nv,:, i1 );    
    P1I = [P1; eye(nw) zeros(nw,nd)];
    P2 = P(nv+1:nv+ny,:, i1 );
    P3 = P(nv+ny+1:end,:, i1 );
    for i2 = 1:nbasis
        Ffr(:,:,i1) = Ffr(:,:,i1) + y{i2}*Fbasisfr{i2}(:,:, i1 );
    end    
    
    FPI = [-Ffr(:,:,i1)*P2+P3; zeros(nd,nw) eye(nd)];
            
    IQCLMI = P1I'*PI(:,:,i1)*P1I + FPI'*M*FPI;
    ev(i1) = max(real(eig(IQCLMI)));           
end

% For debugging
% [LHS,RHS] = showlmi(evallmi(lmisys,xopt),1);

% if max(ev(cidx))>0
%     keyboard
% end





