function [gain,Fopt,PIopt] = robfiltlb(Punc,nmeas,omega)
% ROBFILTLB Lower bound on optimal L2-L2 gain for robust filter design
%
% DESCRIPTION
%   Consider the uncertain system Punc with input d and outputs [y;z].
%   The robust filter design problem is to construct a filter F that uses
%   measurements y to estimate z, i.e. zhat = F*y.  The estimation error
%   is e:=z-zhat. The H_infty robust filter design problem is to construct
%   F that minimizes the "worst-case" L2 gain from d to e.  This function
%   computes a lower bound on the performance achieved by the optimal
%   filter.   The lower bound is computed by solving an LMI for the 
%   optimal, constant filter at each frequency in a frequency grid 
%   (ignoring performance at other frequencies).  
%
%   XXX - Currently the code only works for standard MUSSV blocks 
%   (rep. real, rep. complex, and full complex blocks). 
%   Rep. full and  ultidyn/IQC blocks are currently not allowed. 
%
% INPUTS
%   Punc: (Ny+Nz)-by-Nd Uncertain state space system (USS). The first Ny 
%      outputs of Punc are measurements available to the filter. The last 
%      Nz outputs are signals to be estimated by the filter.
%   Ny: Number of measurements available to the filter
%   omega: Nomeg-by-1 frequency grid on which to compute the lower bound
%
% OUTPUTS
%   gain: Nomeg-by-1 vector of the gains at each frequency. gain(i) is the
%      gain achieved by the optimal constant filter at frequency i.
%      max(gain) is a lower bound on the optimal L2-L2 gain for the
%      H_infty robust filter design problem.
%   Fopt: Nz-by-Ny-by-Nomeg array of complex, constant matrix filters.
%      Fopt(:,:,i) is the optimal filter for frequency i.
%   PIopt:  (Nv+Nw)-by-(Nv+Nw)-by-Nomeg array of IQC scalings.
%      where Nw-by-Nv denotes the dimensions of the uncertainty. The IQC
%      stability condition is satisfied by PIopt and Fopt at each omega.
%
% SYNTAX
%   [gain,Fopt,PIopt] = robfiltlb(Punc,Ny,omega)

% PJS 6/16/2010  Initial Coding

% Get dimensions:   [y;z] = Punc*d
szPunc = size(Punc);
nd = szPunc(2);
ny = nmeas;
nz = szPunc(1)-ny;
ne = nz;

% Get UFRD for generalized plant 
nomeg = length(omega);
Puncfr = ufrd(Punc,omega);

% Get underlying LFT information
%   [v;y;z] = P*[w;d]
%   w = Delta*v
P = lftdata(Puncfr);
P = P.ResponseData; 
blk = getmutoolsblkstruct(Puncfr);
nblk = size(blk,1);

szP = size(P);
nv = szP(1)-szPunc(1);
nw = szP(2)-szPunc(2);

% Construct structure of LMI variables for uncertainty scalings
setlmis([]);
vars = 0;
PI11r = [];  PI11i = [];
PI12r = [];  PI12i = [];
PI22r = [];  PI22i = [];
for i2=1:nblk
    if blk(i2,2) == 0 % repeated real & repeated complex
        blksz = abs(blk(i2,1));
        Xr  = symdec(blksz,vars);
        vars = vars+blksz*(blksz+1)/2;
        Xi = skewdec(blksz,vars);
        vars = vars+blksz*(blksz-1)/2;
        
        if blk(i2,1)<0  % repeated real
            Yr = skewdec(blksz,vars);
            vars = vars+blksz*(blksz-1)/2;
            Yi  = symdec(blksz,vars);
            vars = vars+blksz*(blksz+1)/2;
        else % repeated complex
            Yr = zeros(blksz);
            Yi = zeros(blksz);
        end
        PI11r = blkdiag(PI11r,Xr);
        PI11i = blkdiag(PI11i,Xi);
        
        PI22r = blkdiag(PI22r,-Xr);
        PI22i = blkdiag(PI22i,-Xi);
        
        PI12r = blkdiag(PI12r,Yr);
        PI12i = blkdiag(PI12i,Yi);
        
    else   % full block
        blksz = blk(i2,:);
        X1r = (vars+1)*eye(blksz(1));
        X1i = zeros(blksz(1));
        
        X2r = (vars+1)*eye(blksz(2));
        X2i = zeros(blksz(2));
        
        Yr = zeros( blksz(2), blksz(1) );
        Yi = zeros( blksz(2), blksz(1) );
        
        PI11r = blkdiag(PI11r,X2r);
        PI11i = blkdiag(PI11i,X2i);

        PI22r = blkdiag(PI22r,-X1r);
        PI22i = blkdiag(PI22i,-X1i);
                
        PI12r = blkdiag(PI12r,Yr);
        PI12i = blkdiag(PI12i,Yi);
        
        vars = vars+1;
    end
end

% Create variables for LMI scalings, filter, and gain
PIr = lmivar(3, [PI11r PI12r; PI12r' PI22r] );
PIi = lmivar(3, [PI11i PI12i; -PI12i' PI22i] );
PI11 = lmivar( 3 , [PI11r PI11i; -PI11i PI11r] );

Fr = lmivar(3, reshape( vars+(1:nz*ny) , [nz,ny]));
vars = vars+nz*ny;

Fi = lmivar(3, reshape( vars+(1:nz*ny) , [nz,ny]));
vars = vars+nz*ny;

gsq = lmivar(3, blkdiag(zeros(nw),(vars+1)*eye(nd)) );
vars = vars+1;

lmisys0 = getlmis;

% Solve for lower bound on optimal performance at each frequency
gain = zeros(nomeg,1);
Fopt = zeros(nz,ny,nomeg);
PIopt = zeros(nv+nw,nv+nw,nomeg);
for i1=1:nomeg
    % Get blocks of P
    P1 = P(1:nv,:,i1);
    P1I = [P1; eye(nw) zeros(nw,nd)];
    P2 = P(nv+1:nv+ny,:,i1);
    P3 = P(nv+ny+1:end,:,i1);
        
    % Construct main LMI for constant filter synthesis
    setlmis(lmisys0);
    
    % REAL(L) in (1,1)
    lmiterm([1 1 1 PIr],real(P1I)',real(P1I));
    lmiterm([1 1 1 PIr],imag(P1I)',imag(P1I));
    if ~isempty(PIi)
        lmiterm([1 1 1 PIi],imag(P1I)',real(P1I),'s');    
    end
    lmiterm([-1 1 1 gsq],1,1);
    
    lmiterm([1 2 1 Fr],1,real(P2));
    lmiterm([1 2 1 Fi],1,-imag(P2));
    lmiterm([-1 2 1 0],real(P3));
    
    lmiterm([1 2 2 0],-eye(nz));
    
    % REAL(L) in (2,2)
    lmiblk = [0 2 2 0];
    lmiterm(lmiblk+[1 1 1 PIr],real(P1I)',real(P1I));
    lmiterm(lmiblk+[1 1 1 PIr],imag(P1I)',imag(P1I));
    if ~isempty(PIi)
        lmiterm(lmiblk+[1 1 1 PIi],imag(P1I)',real(P1I),'s');    
    end
    lmiterm(lmiblk+[-1 1 1 gsq],1,1);
    
    lmiterm(lmiblk+[1 2 1 Fr],1,real(P2));
    lmiterm(lmiblk+[1 2 1 Fi],1,-imag(P2));
    lmiterm(lmiblk+[-1 2 1 0],real(P3));
    
    lmiterm(lmiblk+[1 2 2 0],-eye(nz));

    % IMAG(L) in (2,1)
    lmiblk = [0 2 0 0];
    lmiterm(lmiblk+[1 1 1 PIr],real(P1I)',imag(P1I));
    lmiterm(lmiblk+[1 1 1 PIr],-imag(P1I)',real(P1I));
    if ~isempty(PIi)
        lmiterm(lmiblk+[1 1 1 PIi],real(P1I)',real(P1I));
        lmiterm(lmiblk+[1 1 1 PIi],imag(P1I)',imag(P1I));
    end
    
    lmiterm(lmiblk+[1 2 1 Fr],1,imag(P2));
    lmiterm(lmiblk+[1 2 1 Fi],1,real(P2));
    lmiterm(lmiblk+[-1 2 1 0],imag(P3));

    lmiterm(lmiblk+[-1 1 2 -Fr],imag(P2)',1);
    lmiterm(lmiblk+[-1 1 2 -Fi],real(P2)',1);
    lmiterm(lmiblk+[-1 1 2 0],-imag(P3)');
    
    % Add LMI for PI11> 0
    % (Play around with tolerance, xtol*I, if numerical issues arise)
    lmiterm([-2 1 1 PI11],1,1);
    %xtol = 1e-5;       
    %lmiterm([2 1 1 0],xtol*eye(2*nv));
    
    % Solve LMI
    lmisys = getlmis;
    opts = [1e-5 0 0 0 1];
    obj = zeros(vars,1);
    obj(end) = 1;
    [copt,xopt] = mincx(lmisys,obj,opts); 
    
    % Try with Sedumi using the LMILab translator
    %opts.prt = 0;
    %[copt,xopt,Zopt,info] = minsdp('sedumi',lmisys,obj,opts);
    %copt = copt(1);
    
    % Store results
    if ~isempty(xopt)        
        gain(i1) = sqrt(copt);
        i = sqrt(-1);
        
        Fopt(:,:,i1) = dec2mat(lmisys,xopt,Fr) + i*dec2mat(lmisys,xopt,Fi);        
        %PIopt(:,:,i1) = dec2mat(lmisys,xopt,PIr) + i*dec2mat(lmisys,xopt,PIi);
        PIopt(:,:,i1) = dec2mat(lmisys,xopt,PIr);
        if ~isempty(PIi)
            PIopt(:,:,i1) = PIopt(:,:,i1) + i*dec2mat(lmisys,xopt,PIi);
        end
        
        % For debugging: Verify LMI condition is satisfied
        if 0            
            FPI = [-Fopt(:,:,i1)*P2+P3; zeros(nd,nz) eye(nd)];
            M = blkdiag( eye(nz), -gain(i1)^2*eye(nd) );
            
            IQCLMI = P1I'*PIopt(:,:,i1)*P1I + FPI'*M*FPI;
            lmax(i1) = max(real(eig(IQCLMI)));   
            if lmax(i1)> 1e-5
                error('mincx was feasible but LMI was not satisfied');
            end               
        end
    else
        % LMI Infeasible: The LMI should always be feasible so there
        % has been an error if this ever occurs
        gain(i1) = -1;
        error('LMI Infeasible:  This should not happen. Need to debug.');
    end    
end