function [gopt,Fopt,PIopt,xopt] = robfiltsyn(Punc,Fbasis,nmeas,omega)
% ROBFILTSYN Synthesize optimal H-infty robust filter 
%
% DESCRIPTION
%   Consider the uncertain system Punc with input d and outputs [y;z].
%   The robust filter design problem is to construct a filter F that uses
%   measurements y to estimate z, i.e. zhat = F*y.  The estimation error
%   is e:=z-zhat. The H_infty robust filter design problem is to construct
%   F that minimizes the "worst-case" L2 gain from d to e.  This function
%   approximately solves this problem by searching for the optimal
%   filter in the subspace spanned by filter basis functions.
%
%   XXX - Currently the code only works for standard MUSSV blocks 
%   (rep. real, rep. complex, and full complex blocks). 
%   Rep. full and  ultidyn/IQC blocks are currently not allowed. 
%
% INPUTS
%   Punc: (Ny+Nz)-by-Nd Uncertain state space system (USS). The first Ny 
%      outputs of Punc are measurements available to the filter. The last 
%      Nz outputs are signals to be estimated by the filter.
%   Fbasis: Nbasis-by-1 cell array of 1-by-1 TF or SS objects.
%   Ny: Number of measurements available to the filter
%   omega: Nomeg-by-1 frequency grid on which to compute the lower bound
%
% OUTPUTS
%   gain: Optimal worst-case gain achieved by Fopt.
%   Fopt: Nz-by-Ny optimal filter.  Fopt is a linear combination of the
%      basis functions: Fopt = xopt(1)*Fbasis{1}+...+xopt(end)*Fbasis{end}
%   PIopt:  (Nv+Nw)-by-(Nv+Nw)-by-Nomeg array of IQC scalings.
%      where Nw-by-Nv denotes the dimensions of the uncertainty. The IQC
%      stability condition is satisfied by PIopt and Fopt at each omega.
%   xopt: Nbasis-by-1 vector of variables that specify the optimal
%      linear combination of basis functions.
%
% SYNTAX
%   [gopt,Fopt,PIopt,xopt] = robfiltsyn(Punc,Fbasis,omega)

% PJS 6/18/2010  Initial Coding


% Get frequency response of Punc and F basis functions
nomega = length(omega);
Puncfr = ufrd(Punc,omega);

Fbasis = Fbasis(:);
nbasis = length(Fbasis);
Fbasisfr = cell(nbasis,1);
for i1=1:nbasis
    Ftmp = frd(Fbasis{i1},omega);
    Fbasisfr{i1} = Ftmp.ResponseData;
end

% Get underlying LFT information
%   [v;y;z] = P*[w;d]
%   w = Delta*v
Pfr = lftdata(Puncfr);
Pfr = Pfr.ResponseData; 
blk = getmutoolsblkstruct(Puncfr);

% Get dimensions
% XXX Check all bases functions have same dimension?
% XXX Check that dimensions of Puncfr/F are consistent?
szPunc = size(Puncfr);
szF = size(Fbasisfr{1});
szP = size(Pfr);

dim.ny = nmeas;
dim.nz = szPunc(1)-nmeas;
dim.ne = dim.nz;
dim.nd = szPunc(2);
dim.nv = szP(1)-szPunc(1);
dim.nw = szP(2)-szPunc(2);

% Initialize outputs
gopt = [];
Fopt = [];
PIopt = [];
xopt = [];

% Iteration: Synthesize filter using a coarse frequency grid,
% check results on dense grid, and then add frequencies as needed.
% Note: Start coarse grid with the endpoints.
cidx = unique([1 nomega]);

go = 1;
cnt = 0;
evtol = 1e-6; %5e-7; % XXX Tolerance?
while go==1
    % Solve design problem on coarse grid
    cnt = cnt+1;
    [gain,PI,x,ev] = rfsolve(Pfr,Fbasisfr,omega,cidx,blk,dim);
    
    % Check IQC LMI condition on dense grid    
    fidx = find(ev>=evtol);  
    if isempty(fidx)
        % F/PI are feasible on dense grid: Stop
        go = 0;
        newidx = -1;
        
        gopt = gain;
        PIopt = PI;
        xopt = x;
        Fopt = 0;
        for i1=1:nbasis
            Fopt = Fopt+xopt{i1}*Fbasis{i1};
        end
    else
        % Add new frequencies
        % XXX For now, simply add freq corresponding to worst
        %    constraint violation.
        [maxev,newidx] = max(ev);
        if ~isempty(find(cidx==newidx))
            % XXX The worst-case freq is already in our list.
            % If this happens then an error has occurred
            disp('Worst-case freq is already in the constraint list');
            disp('This is a numerical error that needs to be debugged.');
            keyboard
        end
        cidx = [cidx newidx];        
    end
    
    % Diagnostic Information
    fprintf(' cnt = %d gain = %4.3f Nomega = %d Ncidx=%d\t newidx=%d\n',...
        cnt,gain,nomega,length(cidx),newidx);
    
    %figure(1); plot(1:Nomega,ev);title(num2str(cnt)); pause; %keyboard
end

% Convert Fopt to an SS
Fopt = ss(Fopt);

